import numpy as np


def check_units_and_type(input, expected_units, num=None, is_scalar=False):
    """Check whether variable has expected units and type.

    If input does not have units and expected units is not None, then the
    output will be assigned those units. If input has units that conflict
    with expected units a ValueError will be raised.

    Parameters
    ----------
    input : array_like or float
        Variable that will be checked for units and type. Variable should
        be 1D or scalar.
    expected_units : astropy.units or None
        Unit expected for input.
    num : int, optional
        Length expected for input, if it is an array or list.
    is_scalar : bool, optional
        Sets whether the input is a scalar quantity. Default is False for
        array_like inputs; set is_scalar=True to check scalar units only.

    Returns
    ----------
    ndarray or float, with astropy.units
        Returns the input array or scalar with expected units, unless a
        conflict of units or array length occurs, which raise errors.
    """
    if hasattr(input, 'unit'):    # check units
        if expected_units is None:
            raise ValueError('Expecting dimensionless input')
        elif input.unit != expected_units:
            raise ValueError('Expecting input units of ' + str(expected_units))
        else:
            dimensionless = input.value

    else:
        dimensionless = input

    # check its a 1D array and/or convert list to array
    if is_scalar is False:
        dimensionfull = check_array_or_list(dimensionless)
    else:
        dimensionfull = dimensionless

    # include units if appropriate
    if expected_units is not None:
        dimensionfull = dimensionfull * expected_units

    # check array length if appropriate
    if num is not None:
        check_input_size(dimensionfull, num)

    return dimensionfull


def check_array_or_list(input):
    """Return 1D ndarray, if input can be converted and elements are
    non-negative."""
    if type(input) != np.ndarray:
        if type(input) == list:
            output = np.array(input)
        else:
            raise TypeError('Expecting input type as ndarray or list.')
    else:
        output = input

    if output.ndim != 1:
        raise ValueError('Input array must have 1 dimension.')

    if np.sum(output < 0.) > 0:
            raise ValueError("Input array values cannot be negative.")

    return output


def check_input_size(array, num):
    """Check that an array has the expected number of elements."""
    if array.shape[0] != num:
        raise ValueError('Input arrays must have the same length (the \
                          number of clusters).')



import numpy as np
from numpy.testing import assert_equal, assert_allclose, assert_raises
from astropy import units

from clusterlensing.nfw import SurfaceMassDensity


# ----------------------------
# test units of inputs

ncl = 3
r_s = np.repeat(0.1, ncl) * units.Mpc
delta_c = np.repeat(10000., ncl)  # dimensionless
rho_c = np.repeat(0.2, ncl) * units.Msun / units.pc**2 / units.Mpc
sig_off = np.repeat(0.2, ncl) * units.Mpc
rbinarray = np.logspace(np.log10(0.1), np.log10(5.), num=10) * units.Mpc


def test_rs_units():
    rs_unitless = r_s.value
    rs_wrongunits = r_s.value * units.Msun

    # test that a dimensionless rs is coverted to Mpc
    smd = SurfaceMassDensity(rs_unitless, delta_c, rho_c,
                             offsets=sig_off, rbins=rbinarray)
    assert_equal(smd._rs.unit, units.Mpc)

    # test that a non-Mpc unit on rs raises an error
    assert_raises(ValueError, SurfaceMassDensity, rs_wrongunits, delta_c,
                  rho_c, offsets=sig_off, rbins=rbinarray)


def test_rhoc_units():
    rhoc_unitless = rho_c.value
    rhoc_wrongunits = rho_c.value * units.kg / units.m**3

    # test that a dimensionless rho_c is coverted to Msun/Mpc/pc**2
    smd = SurfaceMassDensity(r_s, delta_c, rhoc_unitless,
                             offsets=sig_off, rbins=rbinarray)
    assert_equal(smd._rho_crit.unit, units.Msun / units.Mpc / units.pc**2)

    # test that an incorrect unit on rho_c raises an error
    assert_raises(ValueError, SurfaceMassDensity, r_s, delta_c,
                  rhoc_wrongunits, offsets=sig_off, rbins=rbinarray)


def test_dc_units():
    dc_wrongunits = delta_c * units.Mpc
    assert_raises(ValueError, SurfaceMassDensity, r_s, dc_wrongunits, rho_c,
                  offsets=sig_off, rbins=rbinarray)


# ----------------------------
# test lists as input

rs_list = list(r_s.value)
dc_list = list(delta_c)
rc_list = list(rho_c.value)
soff_list = list(sig_off.value)
rbin_list = list(rbinarray.value)


def test_list_rs():
    smd = SurfaceMassDensity(rs_list, delta_c, rho_c,
                             offsets=sig_off, rbins=rbinarray)
    assert_equal(smd._rs, r_s)


def test_list_dc():
    smd = SurfaceMassDensity(r_s, dc_list, rho_c,
                             offsets=sig_off, rbins=rbinarray)
    assert_equal(smd._delta_c, delta_c)


def test_list_rc():
    smd = SurfaceMassDensity(r_s, delta_c, rc_list,
                             offsets=sig_off, rbins=rbinarray)
    assert_equal(smd._rho_crit, rho_c)


def test_list_sigoff():
    smd = SurfaceMassDensity(r_s, delta_c, rho_c,
                             offsets=soff_list, rbins=rbinarray)
    assert_equal(smd._sigmaoffset, sig_off)


def test_list_rbins():
    smd = SurfaceMassDensity(r_s, delta_c, rho_c,
                             offsets=sig_off, rbins=rbin_list)
    assert_equal(smd._rbins, rbinarray)


def test_input_single_values():
    assert_raises(TypeError, SurfaceMassDensity, r_s[0], delta_c, rho_c,
                  offsets=sig_off, rbins=rbinarray)
    assert_raises(TypeError, SurfaceMassDensity, r_s, delta_c[0], rho_c,
                  offsets=sig_off, rbins=rbinarray)
    assert_raises(TypeError, SurfaceMassDensity, r_s, delta_c, rho_c[0],
                  offsets=sig_off, rbins=rbinarray)
    assert_raises(TypeError, SurfaceMassDensity, r_s, delta_c, rho_c,
                  offsets=sig_off[0], rbins=rbinarray)
    assert_raises(TypeError, SurfaceMassDensity, r_s, delta_c, rho_c,
                  offsets=sig_off, rbins=rbinarray[0])


def test_incompatible_lengths():

    assert_raises(ValueError, SurfaceMassDensity, r_s[0:2], delta_c, rho_c,
                  offsets=sig_off, rbins=rbinarray)

    assert_raises(ValueError, SurfaceMassDensity, r_s, delta_c[0:2], rho_c,
                  offsets=sig_off, rbins=rbinarray)

    assert_raises(ValueError, SurfaceMassDensity, r_s, delta_c, rho_c[0:2],
                  offsets=sig_off[0:2], rbins=rbinarray)

    assert_raises(ValueError, SurfaceMassDensity, r_s, delta_c, rho_c,
                  offsets=sig_off[0:2], rbins=rbinarray)


# ----------------------------
# test NFW centered profiles

toy_data_rbins = np.array([0.1, 0.26591479, 0.70710678,
                           1.88030155, 5.]) * units.Mpc
toy_data_z = np.array([0.05, 1.0])
toy_data_rs = np.array([[0.06920826, 0.06765531],
                        [0.10535529, 0.100552],
                        [0.4255034, 0.37502254],
                        [0., 0.]]) * units.Mpc
toy_data_dc = np.array([[15993.18343503, 7231.03898592],
                        [12760.73852901, 6142.71245062],
                        [6138.9566454, 3615.01284489],
                        [np.nan, np.nan]])
toy_data_rhoc = np.array([[0.13363, 0.4028],
                          [0.13363, 0.4028],
                          [0.13363, 0.4028],
                          [0.13363, 0.4028]]) * (units.Msun / units.Mpc /
                                                 (units.pc**2))

toy_data_sigma = np.array([[[61.676628, 13.91868, 2.4468037,
                             0.37863308, 0.055465539],
                            [79.68557, 17.851227, 3.1250101,
                             0.48266501, 0.070650539]],
                           [[127.43513, 33.316155, 6.4106052,
                             1.035399, 0.15438678],
                            [166.96108, 42.902785, 8.1626417,
                             1.3110978, 0.19504083]],
                           [[878.18174, 390.91924, 120.07696,
                             25.927776, 4.446575],
                            [1260.0668, 536.75538, 156.88796,
                             32.628947, 5.4837611]],
                           [np.empty(5) * np.nan, np.empty(5) * np.nan]])

toy_data_deltasigma = np.array([[[65.309133, 26.372068, 7.6175566,
                                  1.7577702, 0.35356601],
                                 [85.578152, 34.252649, 9.8301535,
                                  2.2592132, 0.45331823]],
                                [[103.74144, 49.563266, 16.319755,
                                  4.0942593, 0.8663956],
                                 [139.95018, 65.609361, 21.269033,
                                  5.2818827, 1.1108229]],
                                [[314.79661, 245.51579, 138.66486,
                                  53.056979, 14.742129],
                                 [483.0392, 364.3582, 195.7899,
                                  71.458238, 19.210324]],
                                [np.empty(5) * np.nan, np.empty(5) * np.nan]])


def test_centered_profiles():

    def _check_sigma(i):
        assert_allclose(sigma_py.value, toy_data_sigma[i],
                        rtol=1e-04)

    def _check_deltasigma(i):
        assert_allclose(deltasigma_py.value, toy_data_deltasigma[i],
                        rtol=1e-04)
        # note: default tolerance is (rtol=1e-07, atol=0)

    zipped_inputs = zip(toy_data_rs, toy_data_dc, toy_data_rhoc)

    # check all 4 sets of toy_data:
    for i, (r_s, delta_c, rho_c) in enumerate(zipped_inputs):

        smd = SurfaceMassDensity(r_s, delta_c, rho_c, rbins=toy_data_rbins)
        sigma_py = smd.sigma_nfw()
        _check_sigma(i)

        deltasigma_py = smd.deltasigma_nfw()
        _check_deltasigma(i)


# ----------------------------
# test NFW offset profiles

toy_data_offset = np.array([0.1, 0.1]) * units.Mpc

toy_data_sigma_off = np.array([[[56.922836, 17.461102, 2.5331495,
                                 0.3805162, 0.055490059],
                                [73.797572, 22.464487, 3.2357614,
                                 0.48506959, 0.070681805]],
                               [[111.72633, 39.527318, 6.6159646,
                                 1.0403518, 0.15445337],
                                [147.04471, 51.205603, 8.427435,
                                 1.3174017, 0.19512522]],
                               [[748.5117, 407.12286, 121.99994,
                                 26.017806, 4.4481085],
                                [1070.3932, 563.12414, 159.64883,
                                 32.747803, 5.4857215]],
                               [np.empty(5) * np.nan, np.empty(5) * np.nan]])

toy_data_deltasigma_off = np.array([[[8.3569013, 16.021837, 7.3340382,
                                      1.7247532, 0.34935389],
                                     [10.902221, 20.846143, 9.4631887,
                                      2.2162585, 0.44782718]],
                                    [[14.476809, 29.198514, 15.730196,
                                      4.0331203, 0.85895897],
                                     [19.335676, 38.775763, 20.500251,
                                      5.2009726, 1.1009191]],
                                    [[57.800595, 136.55704, 131.88898,
                                      52.590289, 14.699197],
                                     [86.975707, 203.06399, 186.61751,
                                      70.815764, 19.149708]],
                                    [np.empty(5) * np.nan,
                                     np.empty(5) * np.nan]])


def test_miscentered_profiles():

    def _check_sigma(i):
        assert_allclose(sigma_py_off.value, toy_data_sigma_off[i],
                        rtol=1e-04)

    def _check_deltasigma(i):
        assert_allclose(deltasigma_py_off.value, toy_data_deltasigma_off[i],
                        rtol=1e-04)

    zipped_inputs = zip(toy_data_rs, toy_data_dc, toy_data_rhoc)

    # check all 4 sets of toy_data:
    for i, (r_s, delta_c, rho_c) in enumerate(zipped_inputs):

        smd = SurfaceMassDensity(r_s, delta_c, rho_c,
                                 offsets=toy_data_offset,
                                 rbins=toy_data_rbins)
        sigma_py_off = smd.sigma_nfw()
        _check_sigma(i)

        deltasigma_py_off = smd.deltasigma_nfw()
        _check_deltasigma(i)



"""NFW profiles for shear and magnification.

Surface mass density and differential surface mass density calculations
for NFW dark matter halos, with and without the effects of miscentering
offsets.
"""

from __future__ import absolute_import, division, print_function

import numpy as np
from astropy import units
from scipy.integrate import simps

from clusterlensing import utils


def _set_dimensionless_radius(self, radii=None, integration=False):
    if radii is None:
        radii = self._rbins  # default radii

    # calculate x = radii / rs
    if integration is True:
        # radii is a 3D matrix (r_eq13 <-> numTh,numRoff,numRbins)
        d0, d1, d2 = radii.shape[0], radii.shape[1], radii.shape[2]

        # with each cluster's rs, we now have a 4D matrix:
        radii_4D = radii.reshape(d0, d1, d2, 1)
        rs_4D = self._rs.reshape(1, 1, 1, self._nlens)
        x = radii_4D / rs_4D

    else:
        # 1D array of radii and clusters, reshape & broadcast
        radii_repeated = radii.reshape(1, self._nbins)
        rs_repeated = self._rs.reshape(self._nlens, 1)
        x = radii_repeated / rs_repeated

    x = x.value
    if 0. in x:
        x[np.where(x == 0.)] = 1.e-10  # hack to avoid infs in sigma

    # dimensionless radius
    self._x = x

    # set the 3 cases of dimensionless radius x
    self._x_small = np.where(self._x < 1. - 1.e-6)
    self._x_big = np.where(self._x > 1. + 1.e-6)
    self._x_one = np.where(np.abs(self._x - 1) <= 1.e-6)


class SurfaceMassDensity(object):
    """Calculate NFW profiles for Sigma and Delta-Sigma.

    Parameters
    ----------
    rs : array_like
        Scale radii (in Mpc) for every halo. Should be 1D, optionally with
        astropy.units of Mpc.
    delta_c : array_like
        Characteristic overdensities for every halo. Should be 1D and have
        the same length as rs.
    rho_crit : array_like
        Critical energy density of the universe (in Msun/Mpc/pc^2) at every
        halo z. Should be 1D, optionally with astropy.units of
        Msun/Mpc/(pc**2), and have the same length as rs.
    offsets : array_like, optional
        Width of the Gaussian distribution of miscentering offsets (in
        Mpc), for every cluster halo. Should be 1D, optionally with
        astropy.units of Mpc, and have the same length as rs. (Note: it is
        common to use the same value for every halo, implying that they are
        drawn from the same offset distribution).
    rbins : array_like, optional
        Radial bins (in Mpc) at which profiles will be calculated. Should
        be 1D, optionally with astropy.units of Mpc.

    Other Parameters
    -------------------
    numTh : int, optional
        Number of bins to use for integration over theta, for calculating
        offset profiles (no effect for offsets=None). Default 200.
    numRoff : int, optional
        Number of bins to use for integration over R_off, for calculating
        offset profiles (no effect for offsets=None). Default 200.
    numRinner : int, optional
        Number of bins at r < min(rbins) to use for integration over
        Sigma(<r), for calculating DeltaSigma (no effect for Sigma ever,
        and no effect for DeltaSigma if offsets=None). Default 20.
    factorRouter : int, optional
        Factor increase over number of rbins, at min(r) < r < max(r), of
        bins that will be used at for integration over Sigma(<r), for
        calculating DeltaSigma (no effect for Sigma, and no effect for
        DeltaSigma if offsets=None). Default 3.

    Methods
    ----------
    sigma_nfw()
        Calculate surface mass density Sigma.
    deltasigma_nfw()
        Calculate differential surface mass density DeltaSigma.

    See Also
    ----------
    ClusterEnsemble : Parameters and profiles for a sample of clusters.
        This class provides an interface to SurfaceMassDensity, and tracks
        a DataFrame of parameters as well as nfw profiles for many clusters
        at once, only requiring the user to specify cluster z and richness,
        at a minimum.

    References
    ----------
    Sigma and DeltaSigma are calculated using the formulas given in:

    C.O. Wright and T.G. Brainerd, "Gravitational Lensing by NFW Halos,"
    The Astrophysical Journal, Volume 534, Issue 1, pp. 34-40 (2000).

    The offset profiles are calculated using formulas given, e.g., in
    Equations 11-15 of:

    J. Ford, L. Van Waerbeke, M. Milkeraitis, et al., "CFHTLenS: a weak
    lensing shear analysis of the 3D-Matched-Filter galaxy clusters,"
    Monthly Notices of the Royal Astronomical Society, Volume 447, Issue 2,
    p.1304-1318 (2015).
    """
    def __init__(self, rs, delta_c, rho_crit, offsets=None, rbins=None,
                 numTh=200, numRoff=200, numRinner=20, factorRouter=3):
        if rbins is None:
            rmin, rmax = 0.1, 5.
            rbins = np.logspace(np.log10(rmin), np.log10(rmax), num=50)
            self._rbins = rbins * units.Mpc
        else:
            # check rbins input units & type
            self._rbins = utils.check_units_and_type(rbins, units.Mpc)

        # check input units & types
        self._rs = utils.check_units_and_type(rs, units.Mpc)
        self._delta_c = utils.check_units_and_type(delta_c, None)
        self._rho_crit = utils.check_units_and_type(rho_crit, units.Msun /
                                                    units.Mpc / (units.pc**2))

        self._numRoff = utils.check_units_and_type(numRoff, None,
                                                   is_scalar=True)
        self._numTh = utils.check_units_and_type(numTh, None, is_scalar=True)
        self._numRinner = utils.check_units_and_type(numRinner, None,
                                                     is_scalar=True)
        self._factorRouter = utils.check_units_and_type(factorRouter, None,
                                                        is_scalar=True)
        # check numbers of bins are all positive
        if (numRoff <= 0) or (numTh <= 0) or (numRinner <= 0) or \
           (factorRouter <= 0):
                raise ValueError('Require numbers of bins > 0')

        self._nbins = self._rbins.shape[0]
        self._nlens = self._rs.shape[0]

        if offsets is not None:
            self._sigmaoffset = utils.check_units_and_type(offsets, units.Mpc)
            utils.check_input_size(self._sigmaoffset, self._nlens)
        else:
            self._sigmaoffset = offsets

        # check array sizes are compatible
        utils.check_input_size(self._rs, self._nlens)
        utils.check_input_size(self._delta_c, self._nlens)
        utils.check_input_size(self._rho_crit, self._nlens)
        utils.check_input_size(self._rbins, self._nbins)

        rs_dc_rcrit = self._rs * self._delta_c * self._rho_crit
        self._rs_dc_rcrit = rs_dc_rcrit.reshape(self._nlens,
                                                1).repeat(self._nbins, 1)

        # set self._x, self._x_big, self._x_small, self._x_one
        _set_dimensionless_radius(self)

    def sigma_nfw(self):
        """Calculate NFW surface mass density profile.

        Generate the surface mass density profiles of each cluster halo,
        assuming a spherical NFW model. Optionally includes the effect of
        cluster miscentering offsets, if the parent object was initialized
        with offsets.

        Returns
        ----------
        Quantity
            Surface mass density profiles (ndarray, in astropy.units of
            Msun/pc/pc). Each row corresponds to a single cluster halo.
        """
        def _centered_sigma(self):
            # perfectly centered cluster case

            # calculate f
            bigF = np.zeros_like(self._x)
            f = np.zeros_like(self._x)

            numerator_arg = ((1. / self._x[self._x_small]) +
                             np.sqrt((1. / (self._x[self._x_small]**2)) - 1.))
            denominator = np.sqrt(1. - (self._x[self._x_small]**2))
            bigF[self._x_small] = np.log(numerator_arg) / denominator

            bigF[self._x_big] = (np.arccos(1. / self._x[self._x_big]) /
                                 np.sqrt(self._x[self._x_big]**2 - 1.))

            f = (1. - bigF) / (self._x**2 - 1.)
            f[self._x_one] = 1. / 3.
            if np.isnan(np.sum(f)) or np.isinf(np.sum(f)):
                print('\nERROR: f is not all real\n')

            # calculate & return centered profiles
            if f.ndim == 2:
                sigma = 2. * self._rs_dc_rcrit * f
            else:
                rs_dc_rcrit_4D = self._rs_dc_rcrit.T.reshape(1, 1,
                                                             f.shape[2],
                                                             f.shape[3])
                sigma = 2. * rs_dc_rcrit_4D * f

            return sigma

        def _offset_sigma(self):

            # size of "x" arrays to integrate over
            numRoff = self._numRoff
            numTh = self._numTh

            numRbins = self._nbins
            maxsig = self._sigmaoffset.value.max()

            # inner/outer bin edges
            roff_1D = np.linspace(0., 4. * maxsig, numRoff)
            theta_1D = np.linspace(0., 2. * np.pi, numTh)
            rMpc_1D = self._rbins.value

            # reshape for broadcasting: (numTh,numRoff,numRbins)
            theta = theta_1D.reshape(numTh, 1, 1)
            roff = roff_1D.reshape(1, numRoff, 1)
            rMpc = rMpc_1D.reshape(1, 1, numRbins)

            r_eq13 = np.sqrt(rMpc ** 2 + roff ** 2 -
                             2. * rMpc * roff * np.cos(theta))

            # 3D array r_eq13 -> 4D dimensionless radius (nlens)
            _set_dimensionless_radius(self, radii=r_eq13, integration=True)

            sigma = _centered_sigma(self)
            inner_integrand = sigma.value / (2. * np.pi)

            # INTEGRATE OVER theta
            sigma_of_RgivenRoff = simps(inner_integrand, x=theta_1D, axis=0,
                                        even='first')

            # theta is gone, now dimensions are: (numRoff,numRbins,nlens)
            sig_off_3D = self._sigmaoffset.value.reshape(1, 1, self._nlens)
            roff_v2 = roff_1D.reshape(numRoff, 1, 1)
            PofRoff = (roff_v2 / (sig_off_3D**2) *
                       np.exp(-0.5 * (roff_v2 / sig_off_3D)**2))

            dbl_integrand = sigma_of_RgivenRoff * PofRoff

            # INTEGRATE OVER Roff
            # (integration axis=0 after theta is gone).
            sigma_smoothed = simps(dbl_integrand, x=roff_1D, axis=0,
                                   even='first')

            # reset _x to correspond to input rbins (default)
            _set_dimensionless_radius(self)

            sigma_sm = np.array(sigma_smoothed.T) * units.solMass / units.pc**2

            return sigma_sm

        if self._sigmaoffset is None:
            finalsigma = _centered_sigma(self)
        elif np.abs(self._sigmaoffset).sum() == 0:
            finalsigma = _centered_sigma(self)
        else:
            finalsigma = _offset_sigma(self)
            self._sigma_sm = finalsigma

        return finalsigma

    def deltasigma_nfw(self):
        """Calculate NFW differential surface mass density profile.

        Generate the surface mass density profiles of each cluster halo,
        assuming a spherical NFW model. Currently calculates centered
        profiles ONLY; DOES NOT have the miscentering implemented.

        Returns
        ----------
        Quantity
            Differential surface mass density profiles (ndarray, in
            astropy.units of Msun/pc/pc). Each row corresponds to a single
            cluster halo.
        """
        def _centered_dsigma(self):
            # calculate g

            firstpart = np.zeros_like(self._x)
            secondpart = np.zeros_like(self._x)
            g = np.zeros_like(self._x)

            small_1a = 4. / self._x[self._x_small]**2
            small_1b = 2. / (self._x[self._x_small]**2 - 1.)
            small_1c = np.sqrt(1. - self._x[self._x_small]**2)
            firstpart[self._x_small] = (small_1a + small_1b) / small_1c

            big_1a = 8. / (self._x[self._x_big]**2 *
                           np.sqrt(self._x[self._x_big]**2 - 1.))
            big_1b = 4. / ((self._x[self._x_big]**2 - 1.)**1.5)
            firstpart[self._x_big] = big_1a + big_1b

            small_2a = np.sqrt((1. - self._x[self._x_small]) /
                               (1. + self._x[self._x_small]))
            secondpart[self._x_small] = np.log((1. + small_2a) /
                                               (1. - small_2a))

            big_2a = self._x[self._x_big] - 1.
            big_2b = 1. + self._x[self._x_big]
            secondpart[self._x_big] = np.arctan(np.sqrt(big_2a / big_2b))

            both_3a = (4. / (self._x**2)) * np.log(self._x / 2.)
            both_3b = 2. / (self._x**2 - 1.)
            g = firstpart * secondpart + both_3a - both_3b

            g[self._x_one] = (10. / 3.) + 4. * np.log(0.5)

            if np.isnan(np.sum(g)) or np.isinf(np.sum(g)):
                print('\nERROR: g is not all real\n', g)

            # calculate & return centered profile
            deltasigma = self._rs_dc_rcrit * g

            return deltasigma

        def _offset_dsigma(self):
            original_rbins = self._rbins.value

            # if offset sigma was already calculated, use it!
            try:
                sigma_sm_rbins = self._sigma_sm
            except AttributeError:
                sigma_sm_rbins = self.sigma_nfw()

            innermost_sampling = 1.e-10  # stable for anything below 1e-5
            inner_prec = self._numRinner
            r_inner = np.linspace(innermost_sampling,
                                  original_rbins.min(),
                                  endpoint=False, num=inner_prec)
            outer_prec = self._factorRouter * self._nbins
            r_outer = np.linspace(original_rbins.min(),
                                  original_rbins.max(),
                                  endpoint=False, num=outer_prec + 1)[1:]
            r_ext_unordered = np.hstack([r_inner, r_outer, original_rbins])
            r_extended = np.sort(r_ext_unordered)

            # set temporary extended rbins, nbins, x, rs_dc_rcrit array
            self._rbins = r_extended * units.Mpc
            self._nbins = self._rbins.shape[0]
            _set_dimensionless_radius(self)  # uses _rbins, _nlens
            rs_dc_rcrit = self._rs * self._delta_c * self._rho_crit
            self._rs_dc_rcrit = rs_dc_rcrit.reshape(self._nlens,
                                                    1).repeat(self._nbins, 1)

            sigma_sm_extended = self.sigma_nfw()
            mean_inside_sigma_sm = np.zeros([self._nlens,
                                             original_rbins.shape[0]])

            for i, r in enumerate(original_rbins):
                index_of_rbin = np.where(r_extended == r)[0][0]
                x = r_extended[0:index_of_rbin + 1]
                y = sigma_sm_extended[:, 0:index_of_rbin + 1] * x

                integral = simps(y, x=x, axis=-1, even='first')

                # average of sigma_sm at r < rbin
                mean_inside_sigma_sm[:, i] = (2. / r**2) * integral

            mean_inside_sigma_sm = mean_inside_sigma_sm * (units.Msun /
                                                           units.pc**2)

            # reset original rbins, nbins, x
            self._rbins = original_rbins * units.Mpc
            self._nbins = self._rbins.shape[0]
            _set_dimensionless_radius(self)
            rs_dc_rcrit = self._rs * self._delta_c * self._rho_crit
            self._rs_dc_rcrit = rs_dc_rcrit.reshape(self._nlens,
                                                    1).repeat(self._nbins, 1)
            self._sigma_sm = sigma_sm_rbins  # reset to original sigma_sm

            dsigma_sm = mean_inside_sigma_sm - sigma_sm_rbins

            return dsigma_sm

        if self._sigmaoffset is None:
            finaldeltasigma = _centered_dsigma(self)
        elif np.abs(self._sigmaoffset).sum() == 0:
            finaldeltasigma = _centered_dsigma(self)
        else:
            finaldeltasigma = _offset_dsigma(self)

        return finaldeltasigma


# Notes on Integration
# ------------------------
# Among the choices for numerical integration algorithms, the
# following options were considered:
# (1) scipy.integrate.dblquad is the obvious choice, but is far too
# slow because it makes of order 10^5 calls to the function to be
# integrated, even for generous settings of epsabs, epsrel. Likely
# it is getting stuck in non-smooth portions of the function space.
# (2) scipy.integrate.simps is fast and converges faster than the
# midpoint integration for both the integration over Roff and theta.
# (3) scipy.integrate.romb was somewhat slower than simps and as well
# as the midpoint rule integration.
# (4) midpoint rule integration via a Riemann Sum (the choice used in
# the previous reincarnation of this project in the C programming
# language) is about the same speed as simps, and converges smoothly
# for both integrals, but requires a much larger number of bins to
# converge to the best estimate.
# (5) numpy.trapz underestimates concave down functions.



"""Galaxy Cluster Ensemble Calculations.

Cluster mass-richness and mass-concentration scaling relations, and NFW
halo profiles for weak lensing shear and magnification, including the
effects of cluster miscentering offsets.

This framework calculates properties and profiles for every individual
cluster, storing the data in tabular form. This is useful for fitting
measured stacked weak lensing profiles, e.g. when you want to account for
the full redshift, mass, and/or centroid offset distributions, and avoid
fitting a single average mass at a single effective redshift.
"""

from __future__ import absolute_import, division, print_function

import numpy as np
import pandas as pd
from astropy import units

from clusterlensing.nfw import SurfaceMassDensity
from clusterlensing import cofm
from clusterlensing import utils

try:
    from IPython.display import display
    notebook_display = True
except:
    notebook_display = False


def calc_delta_c(c200):
    """Calculate characteristic overdensity from concentration.

    Parameters
    ----------
    c200 : ndarray or float
        Cluster concentration parameter.

    Returns
    ----------
    ndarray or float
        Cluster characteristic overdensity, of same type as c200.
    """
    top = (200. / 3.) * c200**3.
    bottom = np.log(1. + c200) - (c200 / (1. + c200))
    return (top / bottom)


def richess_to_mass(richness, norm=2.7e13, slope=1.4):
    """Calculate mass from richness.

    Mass-richness relation assumed is:
    mass = norm * (richness / 20) ^ slope.

    Parameters
    ----------
    richness : ndarray or float
        Cluster richness value(s).
    norm : float, optional
        Normalization of mass-richness relation in units of solar masses,
        defaults to 2.7e13.
    slope : float, optional
        Slope of mass-richness relation in units of solar masses, defaults
        to 1.4.

    Returns
    ----------
    ndarray or float
        Cluster mass(es) in units of solar masses, of same type as
        richness.

    See Also
    ----------
    mass_to_richess : The inverse of this function.
    """
    mass = norm * ((richness / 20.) ** slope)
    return mass


def mass_to_richess(mass, norm=2.7e13, slope=1.4):
    """Calculate richness from mass.

    Mass-richness relation assumed is:
    mass = norm * (richness / 20) ^ slope.

    Parameters
    ----------
    mass : ndarray or float
        Cluster mass value(s) in units of solar masses.
    norm : float, optional
        Normalization of mass-richness relation in units of solar masses,
        defaults to 2.7e13.
    slope : float, optional
        Slope of mass-richness relation in units of solar masses, defaults
        to 1.4.

    Returns
    ----------
    ndarray or float
        Cluster richness(es), of same type as mass.

    See Also
    ----------
    richess_to_mass : The inverse of this function.
    """
    richness = 20. * (mass / norm)**(1. / slope)
    return richness


class ClusterEnsemble(object):
    """Ensemble of galaxy clusters and their properties.

    The ClusterEnsemble object contains parameters and calculated
    values for every individual cluster in a sample. Initializing with
    a collection of redshifts z will fix the number of clusters
    described by the object. Setting the n200 values will then populate
    the object with the full set of available attributes (except for
    the NFW profiles). The values of z, n200, massrich_norm, and
    massrich_slope can be altered and these changes will propogate to
    the other (dependent) attributes.

    In order to generate the attributes containing the NFW halo
    profiles, sigma_nfw and deltasigma_nfw, pass the desired radial
    bins (and additional optional parameters) to the calc_nfw method.

    Parameters
    ----------
    z : array_like
        Redshifts for each cluster in the sample. Should be 1D.

    Attributes
    ----------
    z
    n200
    m200
    c200
    delta_c
    r200
    rs
    Dang_l
    dataframe
    massrich_norm
    massrich_slope
    describe : str
        Short description of the ClusterEnsemble object.
    number : int
        Number of clusters in the sample.
    sigma_nfw : Quantity
        Surface mass density of every cluster, (1D ndarray) with
        astropy.units of Msun/pc/pc. Generated for each element of rbins
        by running calc_nfw(rbins) method.
    deltasigma_nfw : Quantity
        Differential surface mass density of every cluster, (1D ndarray)
        with astropy.units of Msun/pc/pc. Generated for each element of
        rbins by running calc_nfw(rbins) method.

    Methods
    ----------
    calc_nfw(rbins, offsets=None, numTh=200, numRoff=200, numRinner=20,
             factorRouter=3)
        Generate Sigma and DeltaSigma NFW profiles for each cluster,
        optionally with miscentering offsets included.
    show(notebook=True)
        Display table of cluster information and mass-richness
        scaling relaton in use.
    massrich_parameters()
        Print a string showing the mass-richness scaling relation and
        current values of the normalization and slope.

    Other Parameters
    ----------------
    cosmology : str, optional
        Cosmology to use in calculations, default 'Planck13'. Other choices
        are 'WMAP9', 'WMAP7', and 'WMAP5'.
    cm : str, optional
        Concentration-mass relation to use, default 'DuttonMaccio'. Other
        choices are 'Prada' and 'Duffy'.
    """
    def __init__(self, redshifts, cosmology='Planck13', cm='DuttonMaccio'):
        if type(redshifts) != np.ndarray:
            redshifts = np.array(redshifts)
        if redshifts.ndim != 1:
            raise ValueError("Input redshift array must have 1 dimension.")
        if np.sum(redshifts < 0.) > 0:
            raise ValueError("Redshifts cannot be negative.")

        if cosmology == 'Planck13':
            from astropy.cosmology import Planck13 as cosmo
        elif cosmology == 'WMAP9':
            from astropy.cosmology import WMAP9 as cosmo
        elif cosmology == 'WMAP7':
            from astropy.cosmology import WMAP7 as cosmo
        elif cosmology == 'WMAP5':
            from astropy.cosmology import WMAP5 as cosmo
        else:
            raise ValueError('Input cosmology must be one of: \
                              Planck13, WMAP9, WMAP7, WMAP5.')
        self._cosmo = cosmo

        if cm == 'DuttonMaccio':
            self._cm = 'DuttonMaccio'
        elif cm == 'Prada':
            self._cm = 'Prada'
        elif cm == 'Duffy':
            self._cm = 'Duffy'
        else:
            raise ValueError('Input concentration-mass relation must be \
                              one of: DuttonMaccio, Prada, Duffy.')

        self.describe = "Ensemble of galaxy clusters and their properties."
        self.number = redshifts.shape[0]
        self._z = redshifts
        self._rho_crit = cosmo.critical_density(self._z)
        self._massrich_norm = 2.7 * (10**13) * units.Msun
        self._massrich_slope = 1.4
        self._df = pd.DataFrame(self._z, columns=['z'])
        self._Dang_l = cosmo.angular_diameter_distance(self._z)
        self._m200 = None
        self._n200 = None
        self._r200 = None
        self._rs = None
        self._c200 = None
        self._deltac = None

    @property
    def n200(self):
        """Cluster richness values.

        If n200 is set directly, then mass m200 is calculated from n200
        using the mass-richness scaling relation specified by the
        parameters massrich_norm and massrich_slope. If m200 is set
        directly, then n200 is calculated from m200 using the same scaling
        relation. Changes to n200 will propagate to all mass-dependant
        variables.

        :property: Returns cluster richness values
        :property type: ndarray
        :setter: Sets cluster richness values
        :setter type: array_like
        """
        if self._n200 is None:
            raise AttributeError('n200 has not yet been initialized.')
        else:
            return self._n200

    @n200.setter
    def n200(self, richness):
        # Creates/updates values of cluster N200s & dependant variables.
        self._n200 = utils.check_units_and_type(richness, None,
                                                num=self.number)
        self._df['n200'] = pd.Series(self._n200, index=self._df.index)
        self._richness_to_mass()

    @property
    def m200(self):
        """Cluster masses.

        Mass interior to a sphere of radius r200. If m200 is set directly,
        then richness n200 is calculated from m200 using the mass-richness
        scaling relation specified by the parameters massrich_norm and
        massrich_slope. If n200 is set directly, then m200 is calculated
        from n200 using the same scaling relation. Changes to m200 will
        propagate to all mass-dependant variables.

        :property: Returns cluster masses in Msun
        :property type: Quantity
            1D ndarray, with astropy.units of Msun.
        :setter: Sets cluster mass values in Msun
        :setter type: array_like
            Should be 1D array or list, optionally with units.
        """
        if self._m200 is None:
            raise AttributeError('Attribute has not yet been initialized.')
        else:
            return self._m200

    @m200.setter
    def m200(self, mass):
        # Creates/updates values of cluster M200s & dependant variables.
        self._m200 = utils.check_units_and_type(mass, units.Msun,
                                                num=self.number)
        self._df['m200'] = pd.Series(self._m200, index=self._df.index)
        self._mass_to_richness()

    def _richness_to_mass(self):
        # Calculates M_200 for simple power-law scaling relation
        # (with default parameters from arXiv:1409.3571)
        m200 = richess_to_mass(self._n200, norm=self._massrich_norm.value,
                               slope=self._massrich_slope)
        self._m200 = m200 * units.Msun
        self._df['m200'] = pd.Series(self._m200, index=self._df.index)
        self._update_dependant_variables()

    def _mass_to_richness(self):
        # Calculates N_200 for simple power-law scaling relation.
        # Inverse of _richness_to_mass() function.
        n200 = mass_to_richess(self._m200.value,
                               norm=self._massrich_norm.value,
                               slope=self._massrich_slope)
        # note: units cancel but n200 is still a Quantity
        self._n200 = n200
        self._df['n200'] = pd.Series(self._n200, index=self._df.index)
        self._update_dependant_variables()

    @property
    def z(self):
        """Cluster redshifts.

        :property: Returns cluster redshifts
        :property type: ndarray
        :setter: Sets cluster redshifts
        :setter type: array_like
        """
        return self._z

    @z.setter
    def z(self, redshifts):
        # Changes the values of the cluster z's and z-dependant variables.
        self._z = utils.check_units_and_type(redshifts, None, num=self.number)
        self._Dang_l = self._cosmo.angular_diameter_distance(self._z)
        self._df['z'] = pd.Series(self._z, index=self._df.index)
        self._rho_crit = self._cosmo.critical_density(self._z)
        if self._n200 is not None:
            self._update_dependant_variables()

    def _update_dependant_variables(self):
        self._calculate_r200()
        self._calculate_concentrations()
        self._calculate_rs()
        # what else depends on z or m or?

    @property
    def Dang_l(self):
        """Angular diameter distances to clusters.

        :property: Returns distances in Mpc
        :type: Quantity (1D ndarray, with astropy.units of Mpc)
        """
        return self._Dang_l

    @property
    def dataframe(self):
        """Pandas DataFrame of cluster properties.

        :property: Returns DataFrame
        :type: pandas.core.frame.DataFrame
        """
        return self._df

    @property
    def massrich_norm(self):
        """Normalization of Mass-Richness relation:

        M200 = norm * (N200 / 20) ^ slope.

        Changes to massrich_norm will propagate to all mass-dependant
        variables. (This will take current n200 values and convert them to
        m200; in order to retain original values of m200, save them in a
        temporary variable and reset them after this change).

        :property: Returns normalization in Msun
        :property type: Quantity
            float, with astropy.units of Msun. Default is 2.7e+13 Msun.
        :setter: Sets normalization in Msun
        :setter type: float (optionally in astropy.units of Msun)
        """
        return self._massrich_norm

    @massrich_norm.setter
    def massrich_norm(self, norm):
        self._massrich_norm = utils.check_units_and_type(norm, units.Msun,
                                                         is_scalar=True)
        # behavior is to convert current n200 -> new m200
        if hasattr(self, 'n200'):
            self._richness_to_mass()

    @property
    def massrich_slope(self):
        """Slope of Mass-Richness relation:

        M200 = norm * (N200 / 20) ^ slope.

        Changes to massrich_slope will propagate to all mass-dependant
        variables. (This will take current n200 values and convert them to
        m200; in order to retain original values of m200, save them in a
        temporary variable and reset them after this change).

        :property: Returns slope
        :property type: float
            Default value is 1.4.
        :setter: Sets slope
        :setter type: float
        """
        return self._massrich_slope

    @massrich_slope.setter
    def massrich_slope(self, slope):
        if type(slope) == float:
            self._massrich_slope = slope
        else:
            raise TypeError('Expecting input type as float')
        # behavior is to convert current n200 -> new m200
        if hasattr(self, 'n200'):
            self._richness_to_mass()

    def massrich_parameters(self):
        """Print values of M200-N200 scaling relation parameters."""
        print("\nMass-Richness Power Law: M200 = norm * (N200 / 20) ^ slope")
        print("   norm:", self._massrich_norm)
        print("   slope:", self._massrich_slope)

    def show(self, notebook=notebook_display):
        """Display cluster properties and scaling relation parameters."""
        print("\nCluster Ensemble:")
        if notebook is True:
            display(self._df)
        elif notebook is False:
            print(self._df)
        self.massrich_parameters()

    @property
    def r200(self):
        """Cluster Radii.

        r200 is the cluster radius within which the mean density is 200
        times the critical energy density of the universe at that z.

        :property: Returns r200 in Mpc
        :type: Quantity (1D ndarray, in astropy.units of Mpc)
        """
        if self._r200 is None:
            raise AttributeError('Attribute has not yet been initialized.')
        else:
            return self._r200

    @property
    def c200(self):
        """Cluster concentration parameters.

        c200 is calculated from m200 and z using the mass-concentration
        relation specified when ClusterEnsemble object was created (default
        is relation from Dutton & Maccio 2015). Note that c200 = r200/rs.

        :property: Returns c200
        :type: ndarray
        """
        if self._c200 is None:
            raise AttributeError('Attribute has not yet been initialized.')
        else:
            return self._c200

    @property
    def rs(self):
        """Cluster scale radii.

        :property: Returns scale radius in Mpc
        :type: Quantity (1D ndarray, in astropy.units of Mpc)
        """
        if self._rs is None:
            raise AttributeError('Attribute has not yet been initialized.')
        else:
            return self._rs

    @property
    def delta_c(self):
        """Characteristic overdensities of the cluster halos.

        :property: Returns characteristic overdensity
        :type: ndarray
        """
        if self._deltac is None:
            raise AttributeError('Attribute has not yet been initialized.')
        else:
            return self._deltac

    def _calculate_r200(self):
        # calculate r200 from m200
        radius_200 = (3. * self._m200 / (800. * np.pi *
                                         self._rho_crit))**(1. / 3.)
        self._r200 = radius_200.to(units.Mpc)
        self._df['r200'] = pd.Series(self._r200, index=self._df.index)

    def _calculate_concentrations(self):
        if self._cm == 'DuttonMaccio':
            self._c200 = cofm.c_DuttonMaccio(self._z, self._m200,
                                             h=self._cosmo.h)
        elif self._cm == 'Prada':
            self._c200 = cofm.c_Prada(self._z, self._m200, h=self._cosmo.h,
                                      Om_M=self._cosmo.Om0,
                                      Om_L=1 - self._cosmo.Om0)
        elif self._cm == 'Duffy':
            self._c200 = cofm.c_Duffy(self._z, self._m200, h=self._cosmo.h)

        self._df['c200'] = pd.Series(self._c200, index=self._df.index)
        self._calculate_deltac()

    def _calculate_rs(self):
        # cluster scale radius
        self._rs = self._r200 / self._c200
        self._df['rs'] = pd.Series(self._rs, index=self._df.index)

    def _calculate_deltac(self):
        # calculate concentration parameter from c200
        self._deltac = calc_delta_c(self._c200)
        self._df['delta_c'] = pd.Series(self._deltac, index=self._df.index)

    def calc_nfw(self, rbins, offsets=None, numTh=200, numRoff=200,
                 numRinner=20, factorRouter=3):
        """Calculates Sigma and DeltaSigma profiles.

        Generates the surface mass density (sigma_nfw attribute of parent
        object) and differential surface mass density (deltasigma_nfw
        attribute of parent object) profiles of each cluster, assuming a
        spherical NFW model. Optionally includes the effect of cluster
        miscentering offsets.

        Parameters
        ----------
        rbins : array_like
            Radial bins (in Mpc) for calculating cluster profiles. Should
            be 1D, optionally with astropy.units of Mpc.
        offsets : array_like, optional
            Parameter describing the width (in Mpc) of the Gaussian
            distribution of miscentering offsets. Should be 1D, optionally
            with astropy.units of Mpc.

        Other Parameters
        -------------------
        numTh : int, optional
            Parameter to pass to SurfaceMassDensity(). Number of bins to
            use for integration over theta, for calculating offset profiles
            (no effect for offsets=None). Default 200.
        numRoff : int, optional
            Parameter to pass to SurfaceMassDensity(). Number of bins to
            use for integration over R_off, for calculating offset profiles
            (no effect for offsets=None). Default 200.
        numRinner : int, optional
            Parameter to pass to SurfaceMassDensity(). Number of bins at
            r < min(rbins) to use for integration over Sigma(<r), for
            calculating DeltaSigma (no effect for Sigma ever, and no effect
            for DeltaSigma if offsets=None). Default 20.
        factorRouter : int, optional
            Parameter to pass to SurfaceMassDensity(). Factor increase over
            number of rbins, at min(r) < r < max(r), of bins that will be
            used at for integration over Sigma(<r), for calculating
            DeltaSigma (no effect for Sigma, and no effect for DeltaSigma
            if offsets=None). Default 3.
        """
        if offsets is None:
            self._sigoffset = np.zeros(self.number) * units.Mpc
        else:
            self._sigoffset = utils.check_units_and_type(offsets, units.Mpc,
                                                         num=self.number)

        self.rbins = utils.check_units_and_type(rbins, units.Mpc)

        rhoc = self._rho_crit.to(units.Msun / units.pc**2 / units.Mpc)
        smd = SurfaceMassDensity(self.rs, self.delta_c, rhoc,
                                 offsets=self._sigoffset,
                                 rbins=self.rbins,
                                 numTh=numTh,
                                 numRoff=numRoff,
                                 numRinner=numRinner,
                                 factorRouter=factorRouter)

        self.sigma_nfw = smd.sigma_nfw()
        self.deltasigma_nfw = smd.deltasigma_nfw()



from astropy.cosmology import Planck13 as cosmo

# default parameters
h = cosmo.h
Om_M = cosmo.Om0
Om_L = 1. - Om_M


def bias(mass, z, h=h, Om_M=Om_M, Om_L=Om_L):
    """Calculate halo bias, from Seljak & Warren 2004.

    Parameters
    ----------
    mass : ndarray or float
        Halo mass to calculate bias for.
    z : ndarray or float
        Halo z, same type and size as mass.
    h : float, optional
        Hubble parameter, defaults to astropy.cosmology.Planck13.h
    Om_M : float, optional
        Fractional matter density, defaults to
        astropy.cosmology.Planck13.Om0
    Om_L : float, optional
        Fractional dark energy density, defaults to 1-Om_M.

    Returns
    ----------
    ndarray or float
        Returns the halo bias, of same type and size as input mass_halo and
        z_halo. Calculated according to Seljak & Warren 2004 for z = 0.
        For halo z > 0, the non-linear mass is adjusted using the input
        cosmological parameters.

    References
    ----------
    Based on fitting formula derived from simulations in:

    U. Seljak and M.S. Warren, "Large-scale bias and stochasticity of
    haloes and dark matter," Monthly Notices of the Royal Astronomical
    Society, Volume 355, Issue 1, pp. 129-136 (2004).
    """
    M_nl_0 = (8.73 / h) * (10. ** 12.)         # nonlinear mass today [M_sun]
    M_nl = M_nl_0 * (Om_M + Om_L / ((1. + z) ** 3.))  # scaled to z_lens
    x = mass / M_nl
    b = 0.53 + 0.39 * (x ** 0.45) + 0.13 / (40. * x +
                                            1.) + (5.e-4) * (x ** 1.5)
    return b



import numpy as np
from numpy.testing import assert_allclose
from astropy.cosmology import Planck13 as cosmo

from clusterlensing.halobias import bias


def test_Mnl_z0():
    z = 0.
    h = cosmo.h

    def _check_bias_z0(m, ans):
        b = bias(m, z)
        assert_allclose(b, ans)

    # integer multiples of non-linear mass
    M_nl = (8.73 / h) * 10.**12.
    mass = np.array([0., 1., 2., 10., 100.]) * M_nl

    # corresponding bias at z=0
    answer = [0.66, 0.9236707, 1.06577485, 1.64530492, 4.127912607]

    for m, ans in zip(mass, answer):
        yield _check_bias_z0, m, ans



from .clusters import ClusterEnsemble  # noqa
from .nfw import SurfaceMassDensity    # noqa
import cofm                            # noqa
import halobias                        # noqa

__version__ = '0.1.0'                  # noqa



"""Calculate halo concentration from mass and redshift.
"""

from __future__ import absolute_import, division, print_function

import numpy as np
from scipy import integrate
from astropy.cosmology import Planck13 as cosmo

# default cosmological parameters
h = cosmo.h
Om_M = cosmo.Om0
Om_L = 1. - Om_M


def _check_inputs(z, m):
    """Check inputs are arrays of same length or array and a scalar."""
    try:
        nz = len(z)
        z = np.array(z)
    except TypeError:
        z = np.array([z])
        nz = len(z)
    try:
        nm = len(m)
        m = np.array(m)
    except TypeError:
        m = np.array([m])
        nm = len(m)

    if (z < 0).any() or (m < 0).any():
        raise ValueError('z and m must be positive')

    if nz != nm and nz > 1 and nm > 1:
        raise ValueError('z and m arrays must be either equal in length, \
                          OR of different length with one of length 1.')

    else:
        if type(z) != np.ndarray:
            z = np.array(z)
        if type(m) != np.ndarray:
            m = np.array(m)

        return z, m


def c_Prada(z, m, h=h, Om_M=Om_M, Om_L=Om_L):
    """Concentration from c(M) relation published in Prada et al. (2012).

    Parameters
    ----------
    z : float or array_like
        Redshift(s) of halos.
    m : float or array_like
        Mass(es) of halos (m200 definition), in units of solar masses.
    h : float, optional
        Hubble parameter. Default is from Planck13.
    Om_M : float, optional
        Matter density parameter. Default is from Planck13.
    Om_L : float, optional
        Cosmological constant density parameter. Default is from Planck13.

    Returns
    ----------
    ndarray
        Concentration values (c200) for halos.

    Notes
    ----------
    This c(M) relation is somewhat controversial, due to its upturn in
    concentration for high masses (normally we expect concentration to
    decrease with increasing mass). See the reference below for discussion.

    References
    ----------
    Calculation based on results of N-body simulations presented in:

    F. Prada, A.A. Klypin, A.J. Cuesta, J.E. Betancort-Rijo, and J.
    Primack, "Halo concentrations in the standard Lambda cold dark matter
    cosmology," Monthly Notices of the Royal Astronomical Society, Volume
    423, Issue 4, pp. 3018-3030, 2012.
    """

    z, m = _check_inputs(z, m)

    # EQ 13
    x = (1. / (1. + z)) * (Om_L / Om_M)**(1. / 3.)

    # EQ 12
    intEQ12 = np.zeros(len(x))  # integral
    for i in range(len(x)):
        # v is integration variable
        temp = integrate.quad(lambda v: (v / (1 + v**3.))**(1.5), 0, x[i])
        intEQ12[i] = temp[0]

    Da = 2.5 * ((Om_M / Om_L)**(1. / 3.)) * (np.sqrt(1. + x**3.) /
                                             (x**(1.5))) * intEQ12

    # EQ 23
    y = (1.e+12) / (h * m)
    sigma = Da * (16.9 * y**0.41) / (1. + (1.102 * y**0.2) + (6.22 * y**0.333))

    # EQ 21 & 22 (constants)
    c0 = 3.681
    c1 = 5.033
    alpha = 6.948
    x0 = 0.424
    s0 = 1.047  # sigma_0^-1
    s1 = 1.646  # sigma_1^-1
    beta = 7.386
    x1 = 0.526

    # EQ 19 & 20
    cmin = c0 + (c1 - c0) * ((1. / np.pi) * np.arctan(alpha * (x - x0)) + 0.5)
    smin = s0 + (s1 - s0) * ((1. / np.pi) * np.arctan(beta * (x - x1)) + 0.5)

    # EQ 18
    cmin1393 = c0 + (c1 - c0) * ((1. / np.pi) * np.arctan(alpha *
                                                          (1.393 - x0)) + 0.5)
    smin1393 = s0 + (s1 - s0) * ((1. / np.pi) * np.arctan(beta *
                                                          (1.393 - x1)) + 0.5)
    B0 = cmin / cmin1393
    B1 = smin / smin1393

    # EQ 15
    sigma_prime = B1 * sigma

    # EQ 17
    A = 2.881
    b = 1.257
    c = 1.022
    d = 0.06

    # EQ 16
    Cs = A * ((sigma_prime / b)**c + 1.) * np.exp(d / (sigma_prime**2.))

    # EQ 14
    concentration = B0 * Cs

    return concentration


def c_DuttonMaccio(z, m, h=h):
    """Concentration from c(M) relation in Dutton & Maccio (2014).

    Parameters
    ----------
    z : float or array_like
        Redshift(s) of halos.
    m : float or array_like
        Mass(es) of halos (m200 definition), in units of solar masses.
    h : float, optional
        Hubble parameter. Default is from Planck13.

    Returns
    ----------
    ndarray
        Concentration values (c200) for halos.

    References
    ----------
    Calculation from Planck-based results of simulations presented in:

    A.A. Dutton & A.V. Maccio, "Cold dark matter haloes in the Planck era:
    evolution of structural parameters for Einasto and NFW profiles,"
    Monthly Notices of the Royal Astronomical Society, Volume 441, Issue 4,
    p.3359-3374, 2014.
    """

    z, m = _check_inputs(z, m)

    a = 0.52 + 0.385 * np.exp(-0.617 * (z**1.21))  # EQ 10
    b = -0.101 + 0.026 * z                         # EQ 11

    logc200 = a + b * np.log10(m * h / (10.**12))  # EQ 7

    concentration = 10.**logc200

    return concentration


def c_Duffy(z, m, h=h):
    """Concentration from c(M) relation published in Duffy et al. (2008).

    Parameters
    ----------
    z : float or array_like
        Redshift(s) of halos.
    m : float or array_like
        Mass(es) of halos (m200 definition), in units of solar masses.
    h : float, optional
        Hubble parameter. Default is from Planck13.

    Returns
    ----------
    ndarray
        Concentration values (c200) for halos.

    References
    ----------
    Results from N-body simulations using WMAP5 cosmology, presented in:

    A.R. Duffy, J. Schaye, S.T. Kay, and C. Dalla Vecchia, "Dark matter
    halo concentrations in the Wilkinson Microwave Anisotropy Probe year 5
    cosmology," Monthly Notices of the Royal Astronomical Society, Volume
    390, Issue 1, pp. L64-L68, 2008.

    This calculation uses the parameters corresponding to the NFW model,
    the '200' halo definition, and the 'full' sample of halos spanning
    z = 0-2. This means the values of fitted parameters (A,B,C) = (5.71,
    -0.084,-0.47) in Table 1 of Duffy et al. (2008).
    """

    z, m = _check_inputs(z, m)

    M_pivot = 2.e12 / h  # [M_solar]

    A = 5.71
    B = -0.084
    C = -0.47

    concentration = A * ((m / M_pivot)**B) * (1 + z)**C

    return concentration



import numpy as np
from numpy.testing import assert_equal, assert_allclose, assert_raises

from clusterlensing.clusters import ClusterEnsemble


# ------ TOY DATA FOR TESTING --------------
# Note: some of this depends on cosmology & c(M) relation.
# Assuming defaults of Planck13 and DuttonMaccio, respectively.
toy_data_z = np.array([0.05, 1.0])
toy_data_n200 = np.array([[10, 10], [20, 20], [200., 200.], [0., 0.]])
toy_data_m200 = np.array([[1.02310868, 1.02310868],
                          [2.7, 2.7],
                          [67.8209337, 67.8209337],
                          [0., 0.]]) * 10**13
toy_data_r200 = np.array([[0.45043573, 0.31182166],
                          [0.62246294, 0.43091036],
                          [1.82297271, 1.26198329],
                          [0., 0.]])
toy_data_rs = np.array([[0.06898523, 0.06749123],
                        [0.10501577, 0.10030814],
                        [0.42413215, 0.37411303],
                        [0., 0.]])
toy_data_c200 = np.array([[6.52945211, 4.62018029],
                          [5.92732818, 4.29586647],
                          [4.29812433, 3.37326743],
                          [np.inf, np.inf]])
toy_data_deltac = np.array([[16114.78293441, 7270.69406851],
                            [12856.72266431, 6176.07514091],
                            [6183.32001126, 3633.97602181],
                            [np.nan, np.nan]])
# ------------------------------------------


def test_initialization():
    c = ClusterEnsemble(toy_data_z)
    assert_equal(c.z, toy_data_z)
    assert_allclose(c.Dang_l.value, [208.18989, 1697.5794])
    assert_equal(c.number, 2)


def test_update_richness():
    c = ClusterEnsemble(toy_data_z)

    def _check_n_and_m(i):
        assert_equal(c.n200, toy_data_n200[i])
        assert_allclose(c.m200.value, toy_data_m200[i])

    def _check_radii(i):
        assert_allclose(c.r200.value, toy_data_r200[i])
        assert_allclose(c.rs.value, toy_data_rs[i])

    def _check_c200(i):
        assert_allclose(toy_data_c200[i], c.c200)
        if c.c200 is np.real:
            assert_allclose(toy_data_r200[i] / toy_data_rs[i], c.c200)
        else:
            assert(toy_data_r200[i] / toy_data_rs[i] is not np.real)

    def _check_delta_c(i):
        assert_allclose(toy_data_deltac[i], c.delta_c)

    for i, n200 in enumerate(toy_data_n200):
        c.n200 = n200
        yield _check_n_and_m, i
        yield _check_radii, i
        yield _check_c200, i
        yield _check_delta_c, i


def test_update_mass():
    c = ClusterEnsemble(toy_data_z)

    def _check_n_and_m(i):
        assert_equal(c.m200.value, toy_data_m200[i])
        assert_allclose(c.n200, toy_data_n200[i])

    def _check_radii(i):
        assert_allclose(c.r200.value, toy_data_r200[i])
        assert_allclose(c.rs.value, toy_data_rs[i])

    def _check_c200(i):
        assert_allclose(toy_data_c200[i], c.c200)
        if c.c200 is np.real:
            assert_allclose(toy_data_r200[i] / toy_data_rs[i], c.c200)
        else:
            assert(toy_data_r200[i] / toy_data_rs[i] is not np.real)

    def _check_delta_c(i):
        assert_allclose(toy_data_deltac[i], c.delta_c)

    for i, m200 in enumerate(toy_data_m200):
        c.m200 = m200
        yield _check_n_and_m, i
        yield _check_radii, i
        yield _check_c200, i
        yield _check_delta_c, i


def test_negative_z():
    redshifts = np.array([[-1., -999.], [20., 30., -10.]])
    for z in redshifts:
        assert_raises(ValueError, ClusterEnsemble, z)


def test_negative_n200():
    c = ClusterEnsemble(toy_data_z)
    richness = np.array([[-1., -999.], [30., -10.]])

    def set_n200(val):
        c.n200 = val
    for n in richness:
        assert_raises(ValueError, set_n200, n)


def test_wrong_length_richness():
    c = ClusterEnsemble(toy_data_z)
    richness = [np.ones(3), np.arange(4), np.arange(5) + 20.]

    def set_n200(val):
        c.n200 = val
    for n in richness:
        assert_raises(ValueError, set_n200, n)


def test_wrong_length_z():
    c = ClusterEnsemble(toy_data_z)
    redshifts = [np.ones(3), np.arange(4), np.arange(5) + 20.]

    def set_z(val):
        c.z = val
    for z in redshifts:
        assert_raises(ValueError, set_z, z)


def test_wrong_length_update_MNrelation():
    c = ClusterEnsemble(toy_data_z)

    def set_slope(val):
        c.massrich_slope = val

    def set_norm(val):
        c.massrich_norm = val

    assert_raises(TypeError, set_slope, slope=[1.5, 2., 2.5])
    assert_raises(TypeError, set_norm, norm=[1.5e14, 2.e13, 2.5e-2])


def test_update_slope():
    c = ClusterEnsemble(toy_data_z)
    c.n200 = [10, 20]
    slope_before = c._massrich_slope
    slope_after = 2.
    m_before = c.m200
    c.massrich_slope = slope_after
    m_after = c.m200
    assert_equal(m_before[1], m_after[1])
    assert_equal(m_before[0] / m_after[0], (0.5**slope_before) /
                 (0.5**slope_after))


def test_update_norm():
    c = ClusterEnsemble(toy_data_z)
    c.n200 = [10, 20]
    norm_before = c._massrich_norm.value
    norm_after = 2. * norm_before
    m_before = c.m200
    c.massrich_norm = norm_after
    m_after = c.m200
    assert_equal(m_before / m_after, np.array([norm_before / norm_after] * 2))


# ------ TOY DATA FOR TESTING NFW ----------

toy_data_rbins = np.array([0.1, 0.26591479, 0.70710678, 1.88030155, 5.])
toy_data_offset = np.array([0.1, 0.1])

toy_data_sigma = np.array([[[6.16761908e+01, 1.39039946e+01, 2.44275736e+00,
                             3.77903918e-01, 5.53525953e-02],
                            [7.96641862e+01, 1.78324931e+01, 3.12034544e+00,
                             4.81847673e-01, 7.05251489e-02]],
                           [[1.27493931e+02, 3.32915500e+01, 6.40080868e+00,
                             1.03341380e+00, 1.54065429e-01],
                            [1.66972278e+02, 4.28673786e+01, 8.15119054e+00,
                             1.30888896e+00, 1.94689165e-01]],
                           [[879.78399678, 391.2004944, 120.01031327,
                             25.88781944, 4.43731976],
                            [1261.43218563, 536.87349277, 156.77363455,
                             32.58239059, 5.47390998]],
                           [np.empty(5) * np.nan, np.empty(5) * np.nan]])

toy_data_deltasigma = np.array([[[65.43997434, 26.39164262, 7.61618077,
                                  1.75644108, 0.35317515],
                                 [85.68440795, 34.26283038, 9.82631838,
                                  2.25736485, 0.45282956]],
                                [[103.99931225, 49.62147716, 16.32117621,
                                  4.0917027, 0.86548029],
                                 [140.17269195, 65.64905033, 21.26478668,
                                  5.27806563, 1.10966763]],
                                [[315.90515388, 246.17865955, 138.86425678,
                                  53.06822795, 14.73246768],
                                 [484.19549909, 364.98607848, 195.93777776,
                                  71.4483272, 19.19583107]],
                                [np.empty(5) * np.nan, np.empty(5) * np.nan]])

toy_data_sigma_off = np.array([[[5.69501008e+01, 1.74504112e+01, 2.52901319,
                                 3.79783870e-01, 5.53770689e-02],
                                [7.38052696e+01, 2.24484132e+01, 3.23098141,
                                 4.84248607e-01, 7.05563626e-02]],
                               [[111.8118779, 39.51432979, 6.60603604,
                                 1.03835891, 0.15413189],
                                [147.0904175, 51.1793199, 8.41578176,
                                 1.31518391, 0.19477342]],
                               [[749.80727064, 407.48783438, 121.93704135,
                                 25.97782737, 4.43885162],
                                [1071.49027696, 563.33246958, 159.53729442,
                                 32.7011807, 5.47586815]],
                               [np.empty(5) * np.nan, np.empty(5) * np.nan]])

toy_data_deltasigma_off = np.array([[[8.36836538, 16.03773137, 7.33259501,
                                      1.72339183, 0.3489577],
                                     [10.91061883, 20.85627022, 9.45937694,
                                      2.21438965, 0.4473347]],
                                    [[14.50275018, 29.23923747, 15.7315571,
                                      4.03049431, 0.8580319],
                                     [19.35639601, 38.80576556, 20.49612042,
                                      5.19710646, 1.09975454]],
                                    [[57.97626837, 136.93183823, 132.08593043,
                                      52.60120424, 14.68947908],
                                     [87.14987636, 203.42283987, 186.76588425,
                                      70.8056412, 19.13516388]],
                                    [np.empty(5) * np.nan,
                                     np.empty(5) * np.nan]])


def test_nfw_centered():
    c = ClusterEnsemble(toy_data_z)

    def _check_sigma(i, j):
        assert_allclose(c.sigma_nfw[j].value, toy_data_sigma[i, j],
                        rtol=1e-4)

    def _check_deltasigma(i, j):
        assert_allclose(c.deltasigma_nfw[j].value, toy_data_deltasigma[i, j],
                        rtol=1e-4)

    for i, n200 in enumerate(toy_data_n200):
        c.n200 = n200
        c.calc_nfw(toy_data_rbins)
        for j in range(c.z.shape[0]):
            yield _check_sigma, i, j
            yield _check_deltasigma, i, j


def test_nfw_offset():
    c = ClusterEnsemble(toy_data_z)

    def _check_sigma(i, j):
        assert_allclose(c.sigma_nfw[j].value, toy_data_sigma_off[i, j],
                        rtol=10**-4)

    def _check_deltasigma(i, j):
        assert_allclose(c.deltasigma_nfw[j].value,
                        toy_data_deltasigma_off[i, j],
                        rtol=10**-4)

    for i, n200 in enumerate(toy_data_n200[:-1]):
        c.n200 = n200
        c.calc_nfw(toy_data_rbins, offsets=toy_data_offset)
        for j in range(c.z.shape[0]):
            yield _check_sigma, i, j
            yield _check_deltasigma, i, j


# ------------------------------------------


def test_for_infs_in_miscentered_c_calc():
    c = ClusterEnsemble(toy_data_z)

    def _check_sigma_off(arr):
        if np.isnan(arr.sum()):
            raise ValueError('sigma_off result contains NaN', arr)
        if np.isinf(arr.sum()):
            raise ValueError('sigma_off result contains Inf', arr)

    def _check_deltasigma_off(arr):
        if np.isnan(arr.sum()):
            raise ValueError('sigma_off result contains NaN', arr)
        if np.isinf(arr.sum()):
            raise ValueError('sigma_off result contains Inf', arr)

    # last element in toy_data is n200=0 -> NaN (skip for this check)
    for n200 in toy_data_n200[:-1]:
        c.n200 = n200
        c.calc_nfw(toy_data_rbins, offsets=toy_data_offset)
        for i in range(c.z.shape[0]):
            yield _check_sigma_off, c.sigma_nfw[i].value
            yield _check_deltasigma_off, c.deltasigma_nfw[i].value



import os
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

DESCRIPTION = "Galaxy Cluster and Weak Lensing Tools"
LONG_DESCRIPTION = """
cluster-lensing: galaxy cluster halo calculations
======================================================
This package includes tools for calculating a variety of galaxy cluster properties, as well as mass-richness and mass-concentration scaling relations, and weak lensing profiles. These include surface mass density (Sigma) and differential surface mass density (DeltaSigma) for NFW halos, both with and without the effects of cluster miscentering.
For more information, visit http://jesford.github.io/cluster-lensing
"""
NAME = "cluster-lensing"
AUTHOR = "Jes Ford"
AUTHOR_EMAIL = "jesford@uw.edu"
MAINTAINER = "Jes Ford"
MAINTAINER_EMAIL = "jesford@uw.edu"
URL = 'http://github.com/jesford/cluster-lensing'
DOWNLOAD_URL = 'http://github.com/jesford/cluster-lensing'
LICENSE = 'MIT'
VERSION = '0.1.0'

setup(name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    url=URL,
    download_url=DOWNLOAD_URL,
    license=LICENSE,
    packages=['clusterlensing', 'clusterlensing/tests'],
    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7'
        ]
     )
